"""Dorkanoid by Hannu Kankaanpaa, May 11. 2003

The game tries to be fair:
- yellow balls hit the paddle area at rougly constant intervals,
  even though they may bounce several times when going across
  the screen
- red killer balls are thrown at random but they never hit
  the paddle area at the same time and same location when a yellow
  ball does.

Try to alter the game constants that are defined below.
"""

try:
    import psyco
    psyco.full()
except ImportError:
    pass

import math
import random
import os
import time
import pickle

import pygame
from pygame.locals import *


#####################################################################
# Game constants                                                    #
#####################################################################
GRAVITY = 150
PADDLE_Y = 530             # paddle location in y-coodinates
BALLRADIUS = 5
# Lower elasticy and the balls won't bounce so much
# Too low elasticy and they won't make it through the screen and
# the game will hang
ELASTICY = 0.85
WIDTH, HEIGHT = 800, 600
# The area where the balls should drop
AREA_LEFT = 378 - 55/2
AREA_RIGHT = 738 + 55/2
# Where the balls spawn
SPAWN_Y = HEIGHT + 10
# How low is the minimum height the balls must rise
LOWEST_THROW = 400
PADDLE_WIDTH = 55
# How fast are human reflexes at best? (180 ms) The balls
# will never hit ground with shorter intervals than this
REFLEX_MINIMUM = 0.18

#####################################################################

# vy (velocity y) is limited in terms of roof and the lowest throw
MAX_VY_SQR = 2 * (SPAWN_Y - BALLRADIUS) * GRAVITY
MIN_VY_SQR = 2 * (SPAWN_Y - LOWEST_THROW) * GRAVITY

def loadImage(filename, colorkey=None):
    """Load an image from the data-directory."""
    fullname = os.path.join('data', filename)
    image = pygame.image.load(fullname)
    image = image.convert()
    if colorkey:
        image.set_colorkey(colorkey)
    return image


def solveImpact(bx, by, vx, vy):
    """What x, what time, what vertical velocity (vy), as the ball
    impacts at PADDLE_Y - BALLRADIUS?
    """
    root = math.sqrt(2.0*GRAVITY*(PADDLE_Y - BALLRADIUS - by) + vy*vy)
    ix = bx + vx * (root - vy) / GRAVITY
    it = (root - vy) / GRAVITY
    ivy = vy + GRAVITY * it
    return ix, it, ivy


def calculateImpactTimes(ix, iy, ivx, ivy, time):
    """Return an array that contains impact times and x-positions
    in a tuple for given initial coordinates, velocity and time.
       
    Result example: [(2.4, 400), (3.4, 600), (4.6, 750)]
    Which means: The ball will first hit at time 2.4 (seconds)
    at position (400, PADDLE_Y - BALLRADIUS). Then it will hit
    at time 3.4 (seconds) at position (600, PADDLE_Y - BALLRADIUS)
    """
    impactTimes = []
    it_accumulated = time
    while True:
        ix, it, ivy = solveImpact(ix, iy, ivx, ivy)
        if ix > AREA_RIGHT + 10: break
        it_accumulated += it
        iy = PADDLE_Y - BALLRADIUS
        impactTimes.append((it_accumulated, ix))
        ivy *= -ELASTICY
        ivx *= ELASTICY
    return impactTimes


class Ball:
    """This class represents a ball (both yellow and red)."""

    def __init__(self, x, y, vx, vy, time, killer):
        """Creates a new ball.
        
        time -- the creation time
        killer -- tells if the ball is red
        """
        self.x = x
        self.y = y
        self.vx = vx
        self.vy = vy
        self.killer = killer
        if not self.killer:
            self.impactTimes = calculateImpactTimes(x, y, vx, vy, time)
        else:
            ix, it, ivy = solveImpact(x, y, vx, vy)
            self.impactTimes = [(it + time, ix)]

    def move(self, dt, paddle, game):
        """Move the ball.
        
        dt -- time step
        paddle -- player's paddle object
        game -- game object
        """
        oldy = self.y
        self.x += self.vx * dt
        # Use mid-point method for precise gravity.
        # See http://www.niksula.cs.hut.fi/~hkankaan/Homepages/gravity.html
        self.vy += GRAVITY * dt * 0.5
        self.y += self.vy * dt
        self.vy += GRAVITY * dt * 0.5

        # Bouncing from a paddle:
        if self.vy > 0 and \
           oldy + BALLRADIUS <= paddle.y and \
           self.y + BALLRADIUS > paddle.y and \
           abs(self.x - paddle.x) < BALLRADIUS + PADDLE_WIDTH/2:
            # Killers kill instead of bouncing:
            if self.killer:
                game.shouldRun = False
            self.vy *= -ELASTICY
            self.vx *= ELASTICY
            # Adjust y location so that the ball is now as much
            # above the paddle as it was below (not entirely
            # precise but suffices).
            self.y = 2*(paddle.y - BALLRADIUS) - self.y
            game.boingSound.play()

        # If the ball hits the area below paddle, and it's not a killer:
        if not self.killer and \
           self.y > PADDLE_Y + 20 and \
           self.x > AREA_LEFT + 5 and \
           self.x < AREA_RIGHT - 10:
            game.shouldRun = False


    def deletable(self):
        """Can the ball be deleted?"""
        return self.x > WIDTH + BALLRADIUS


def createBall(balls, mindelay, time, killer):
    """Create a new ball.
    
    balls -- a list of other balls. Their impacttimes are
             used to determine how the new ball should be created
    mindelay -- minimum delay between two bounces
    time -- current time
    killer -- are we creating a killer ball?
    """

    def xImpact(at_x, bx, by, vy):
        """Return a vx (velocity x) depending on where the ball
        should collide (at_x), where it begins (bx, by) and what is
        its initial vertical velocity (vy).
        """
        return ((bx - at_x) / (2*(BALLRADIUS + by - PADDLE_Y))) * \
                (vy + math.sqrt(2*GRAVITY*(PADDLE_Y - BALLRADIUS - by)+vy*vy))

    # Randomize velocities and initial x-coordinate.
    vy = -math.sqrt(random.uniform(MIN_VY_SQR, MAX_VY_SQR))
    bx = random.uniform(5, 80)
    vx = random.uniform(xImpact(AREA_LEFT, bx, SPAWN_Y, vy),
                        xImpact(AREA_RIGHT, bx, SPAWN_Y, vy))

    def isKillerTooClose(ix1, ix2, it1, it2):
        """Is a killer ball hitting the floor too closely with
        a normal ball?
        """
        # Balls hitting on corners need special treatment:
        if ix1 < AREA_LEFT + PADDLE_WIDTH/2:
            ix1 = AREA_LEFT + PADDLE_WIDTH/2
        if ix1 > AREA_RIGHT - PADDLE_WIDTH/2:
            ix1 = AREA_RIGHT - PADDLE_WIDTH/2
        # Same time and same place? -> exit
        if abs(ix1 - ix2) < PADDLE_WIDTH/2 + 15 and \
           abs(it1 - it2) < mindelay*2:
            return True
        return False
    

    if killer:
        # Creating a killer ball.
        
        # First, solve when it will impact (a killer ball can't bounce).
        ix, it, ivy = solveImpact(bx, SPAWN_Y, vx, vy)
        it += time
        
        # Then, make sure that none of the other balls (unless killers)
        # will hit the paddle at the same location at the same time.
        # (since that would make the game unfair)
        for ball in balls:
            if not ball.killer:
                for it1, ix1 in ball.impactTimes:
                    if isKillerTooClose(ix1, ix, it1, it):
                        return

        balls.append(Ball(bx, SPAWN_Y, vx, vy, time, killer=True))
    else:
        # Creating a normal ball.
        
        # First, solve all impact times.
        impactTimes = calculateImpactTimes(bx, SPAWN_Y, vx, vy, time)
        
        # Then, make sure that none of the other balls hit the board
        # too closely in time (mindelay) and that some killer ball
        # isn't already hitting that location at the same time.
        for ball in balls:
            for it1, ix1 in ball.impactTimes:
                for it2, ix2 in impactTimes:
                    if ball.killer:
                        if isKillerTooClose(ix1, ix2, it1, it2):
                            return
                    elif abs(it1 - it2) < mindelay:
                        return

        balls.append(Ball(bx, SPAWN_Y, vx, vy, time, killer=False))


class Paddle:
    """Player's paddle."""
    
    def __init__(self, x, y, gfxFile):
        self.x = x
        self.y = y
        self.gfx = loadImage(gfxFile, (0, 0, 0))
        self.width, self.height = self.gfx.get_size()

    def move(self):
        if pygame.mouse.get_focused():
            self.x += pygame.mouse.get_rel()[0]
        self.x = max(self.x, AREA_LEFT + self.width / 2)
        self.x = min(self.x, AREA_RIGHT - self.width / 2)
    

class Dorkanoid:
    """The game itself. Start it with run()."""

    def __init__(self):
        # Basic PyGame initialization:
        pygame.init()
        pygame.mixer.init()
        pygame.mouse.set_visible(0)
        pygame.event.set_grab(1)
        self.screen = pygame.display.set_mode((WIDTH, HEIGHT), FULLSCREEN, 16)
        pygame.display.set_caption('Dorkanoid')
        # Member variables:
        self.backgroundGfx = loadImage('bg.png')
        self.ballGfx = loadImage('ball.png', (0, 0, 0))
        self.killerBallGfx = loadImage('ball2.png', (0, 0, 0))
        self.paddle = Paddle((AREA_LEFT + AREA_RIGHT) / 2, PADDLE_Y, 'paddle.png')
        self.font = pygame.font.Font(None, 24)
        # When exit is set to True, the game will exit.
        self.exit = False
        self.boingSound = pygame.mixer.Sound(os.path.join('data', 'bounce.wav'))
        self.hardness = 1
        # Text dictionary maps text to corresponding bitmap, so that same text
        # can be blitted rapidly.
        self.textDic = {}
        self.loadHighscores()

    def saveHighscores(self):
        f = file('scores', 'w')
        try:
            pickle.dump(self.highscores, f)
        finally:
            f.close()

    def loadHighscores(self):
        try:
            f = file('scores')
            try:
                self.highscores = pickle.load(f)
            finally:
                f.close()
        except IOError:
            self.highscores = [0, 0, 0, 0]

    def blitText(self, text, x, y, color = (200, 200, 255), cache = True):
        """Blit text on the screen (self.screen). If cache is True, the
        text bitmap will be stored into textDic for reuse.
        """
        if cache:
            if not self.textDic.has_key(text):
                self.textDic[text] = self.font.render(text, 1, color)
            self.screen.blit(self.textDic[text], (x, y))
        else:
            textGfx = self.font.render(text, 1, color)
            self.screen.blit(textGfx, (x, y))

    def playOnce(self):
        """Play a single game."""
        self.balls = []
        self.lastTime = self.beginTime = time.clock()
        self.shouldRun = True
        # Delays on different hardness levels.
        delays = (1.0, 0.25, 0.25, 0.10)
        self.mindelay = delays[self.hardness]
        # When is the next red (killer) ball spawned?
        self.nextRedBallTime = random.uniform(0.3, 1.5)
        self.points = 0
        while self.shouldRun:
            # Set delta time (dt)
            dt = float(time.clock() - self.lastTime)
            self.lastTime = time.clock()

            # Adjust points and increase hardness
            self.points += dt * 100
            self.mindelay -= (dt / 70) * self.mindelay
            if self.mindelay < 0: self.mindelay = 0

            # Spawn red balls on hardness level 2
            if self.hardness == 2:
                self.nextRedBallTime -= dt
                if self.nextRedBallTime < 0:
                    createBall(self.balls, self.mindelay + REFLEX_MINIMUM,
                               time.clock(), True)
                    self.nextRedBallTime = random.uniform(0.3, 1.5)
                
            createBall(self.balls, self.mindelay + REFLEX_MINIMUM,
                       time.clock(), False)

            # Move all balls
            for ball in self.balls:
                ball.move(dt, self.paddle, self)

            # Delete balls
            self.balls = filter(lambda x: not x.deletable(), self.balls)

            self.paddle.move()

            # Drawing routines follow:
            self.screen.blit(self.backgroundGfx, (0, 0))
            self.screen.blit(self.paddle.gfx, (self.paddle.x -
                                               self.paddle.width / 2,
                                               self.paddle.y))

            for ball in self.balls:
                self.screen.blit(ball.killer and self.killerBallGfx
                                              or self.ballGfx,
                                 (ball.x - BALLRADIUS / 2,
                                  ball.y - BALLRADIUS / 2))

            self.blitText("Points: " + str(int(self.points)), 10, 10,
                          cache=False)
            pygame.display.flip()

            # Input handing (just exit input here)
            for event in pygame.event.get():
                if event.type == QUIT:
                    self.exit = True
                    self.shoudRun = False
                elif event.type == KEYDOWN and event.key == K_ESCAPE:
                    self.shouldRun = False

    def run(self):
        """Run the game. This function shows the main menu and
        starts playOnce() once the player presses space.
        """
        self.screen.blit(self.backgroundGfx, (0, 0))
        while not self.exit:
            # Draw menu text on the screen
            self.screen.blit(self.backgroundGfx, (200, 50), (200, 50, 300, 220))

            self.blitText("Dorkanoid!", 200, 50, (255, 255, 255))
            self.blitText("Hardness: " +
                          str(self.hardness + 1), 200, 100)
            self.blitText("Press 1,2,3,4 to adjust. 4 is hardest", 200, 130)
            self.blitText("Press space to begin, esc to exit", 200, 170,
                          (255, 255, 200))
            self.blitText("Hi-score: " + str(self.highscores[self.hardness]),
                          200, 220)
            
            pygame.display.flip()

            # Handle input
            for event in pygame.event.get():
                if event.type == QUIT:
                    self.exit = True
                elif event.type == KEYDOWN:
                    if event.key == K_1: self.hardness = 0
                    elif event.key == K_2: self.hardness = 1
                    elif event.key == K_3: self.hardness = 2
                    elif event.key == K_4: self.hardness = 3
                    elif event.key == K_SPACE:
                        self.playOnce()
                        self.points = int(self.points)
                        if self.points > self.highscores[self.hardness]:
                            self.highscores[self.hardness] = self.points
                        self.saveHighscores()
                    elif event.key == K_ESCAPE: self.exit = True

# And finally, the typical start-up row.
if __name__ == '__main__': Dorkanoid().run()

